@@ -31,6 +31,7 @@
 import time
 from abc import ABC, abstractmethod
 from enum import IntEnum
+from typing import Optional
 
 
 class ConnectionStatus(IntEnum):
@@ -108,10 +109,52 @@ class ConnectionStatus(IntEnum):
 """
 
 
+class DatagramPacket:
+
+    def __init__(self, data: bytes, address: tuple):
+        super().__init__()
+        self.__data = data
+        self.__address = address
+
+    @property
+    def data(self) -> bytes:
+        return self.__data
+
+    @property
+    def offset(self) -> int:
+        return 0
+
+    @property
+    def length(self) -> int:
+        return len(self.__data)
+
+    @property
+    def address(self) -> tuple:
+        return self.__address
+
+    @property
+    def ip(self) -> str:
+        return self.__address[0]
+
+    @property
+    def port(self) -> int:
+        return self.__address[1]
+
+
 class Connection:
 
     EXPIRES = 28  # seconds
 
+    """
+        Max count for caching packages
+        ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
+
+        Each UDP data package is limited to no more than 576 bytes, so set the
+        MAX_CACHE_SPACES to about 200,000 means it would take up to 100MB memory
+        for caching in one connection.
+    """
+    MAX_CACHE_SPACES = 1024 * 200
+
     def __init__(self, local_address: tuple, remote_address: tuple):
         super().__init__()
         self.__local_address = local_address
@@ -122,6 +165,8 @@ class Connection:
         now = time.time()
         self.__last_sent_time = now - self.EXPIRES - 1
         self.__last_received_time = now - self.EXPIRES - 1
+        # received packages
+        self.__cargoes = []
 
     @property
     def local_address(self) -> tuple:
@@ -133,27 +178,25 @@ class Connection:
         """ remote ip, port """
         return self.__remote_address
 
-    @property
-    def is_connected(self) -> bool:
-        return ConnectionStatus.is_connected(status=self.get_status())
+    def is_connected(self, now: float) -> bool:
+        status = self.get_status(now=now)
+        return ConnectionStatus.is_connected(status=status)
 
-    @property
-    def is_expired(self) -> bool:
-        return ConnectionStatus.is_expired(status=self.get_status())
+    def is_expired(self, now: float) -> bool:
+        status = self.get_status(now=now)
+        return ConnectionStatus.is_expired(status=status)
 
-    @property
-    def is_error(self) -> bool:
-        return ConnectionStatus.is_error(status=self.get_status())
+    def is_error(self, now: float) -> bool:
+        status = self.get_status(now=now)
+        return ConnectionStatus.is_error(status=status)
 
-    def get_status(self, now: float=0):
+    def get_status(self, now: float):
         """
         Get connection status
 
         :param now: timestamp in seconds
         :return: new status
         """
-        if now < 1:
-            now = time.time()
         # pre-checks
         if now < self.__last_received_time + self.EXPIRES:
             # received response recently
@@ -203,17 +246,33 @@ class Connection:
                 self.__status = ConnectionStatus.Expired
         return self.__status
 
-    def update_sent_time(self, now: float) -> ConnectionStatus:
-        # update last sent time
+    def update_sent_time(self, now: float):
         self.__last_sent_time = now
-        # update and return new status
-        return self.get_status(now=now)
 
-    def update_received_time(self, now: float) -> ConnectionStatus:
-        # update last received time
+    def update_received_time(self, now: float):
         self.__last_received_time = now
-        # update and return new status
-        return self.get_status(now=now)
+
+    def receive(self) -> Optional[DatagramPacket]:
+        """
+        Get received data package from buffer, non-blocked
+
+        :return: received data and source address
+        """
+        if len(self.__cargoes) > 0:
+            return self.__cargoes.pop(0)
+
+    def cache(self, packet: DatagramPacket) -> Optional[DatagramPacket]:
+        # 1. check memory cache status
+        ejected = None
+        if self._is_cache_full(count=len(self.__cargoes)):
+            # drop the first package
+            ejected = self.__cargoes.pop(0)
+        # 2. append the new package to the end
+        self.__cargoes.append(packet)
+        return ejected
+
+    def _is_cache_full(self, count: int) -> bool:
+        return count > self.MAX_CACHE_SPACES
 
 
 class ConnectionHandler(ABC):
